import pytest
from lab_partner.option_set import KeyValueOption, Option, OptionSet, ValueOption


def test_equal_options_replace_each_other():
    option_set = OptionSet()
    option1 = Option('--o')
    option2 = Option('--o')
    id2 = id(option2)

    option_set.put(option1)
    option_set.put(option2)

    assert len(option_set) == 1, 'Only a single option should be in the set'
    assert id2 == id(option_set.to_list().pop()), 'ID of option should equal to option 2'

def test_equal_value_options_replace_each_other():
    option_set = OptionSet()
    option1 = ValueOption('--o', 'dog')
    option2 = ValueOption('--o', 'dog')
    id2 = id(option2)

    option_set.put(option1)
    option_set.put(option2)
    assert len(option_set) == 1, 'Only a single value option should be in the set'
    assert id2 == id(option_set.to_list().pop()), 'ID of option should equal to option 2'

def test_no_duplicate_value_options_different_values():
    option_set = OptionSet()
    option1 = ValueOption('--o', 'dog')
    option2 = ValueOption('--o', 'cat')

    option_set.put(option1)
    option_set.put(option2)
    assert len(option_set) == 1, 'Only a single value option should be in the set'
    assert option2 == option_set.to_list().pop(), 'option in set should equal to option 2'

def test_different_value_options_coexist():
    option_set = OptionSet()
    option1 = ValueOption('--o', 'dog')
    option2 = ValueOption('--p', 'dog')

    option_set.put(option1)
    option_set.put(option2)
    assert option1 in option_set
    assert option2 in option_set

def test_equal_key_value_options_replace_each_other():
    option_set = OptionSet()
    option1 = KeyValueOption('--o', 'dog', 'rover')
    option2 = KeyValueOption('--o', 'dog', 'rover')
    id2 = id(option2)

    option_set.put(option1)
    option_set.put(option2)
    assert len(option_set) == 1, 'Only a single key value option should be in the set'
    assert id2 == id(option_set.to_list().pop()), 'ID of option should equal to option 2'

def test_no_duplicate_key_value_options_different_values():
    option_set = OptionSet()
    option1 = KeyValueOption('--o', 'dog', 'rover')
    option2 = KeyValueOption('--o', 'dog', 'rex')

    option_set.put(option1)
    option_set.put(option2)
    assert len(option_set) == 1, 'Only a single key value option should be in the set'
    assert option2 == option_set.to_list().pop(), 'option in set should equal to option 2'

def test_different_key_value_options_coexist():
    option_set = OptionSet()
    option1 = KeyValueOption('--o', 'dog', 'rover')
    option2 = KeyValueOption('--p', 'dog', 'rover')

    option_set.put(option1)
    option_set.put(option2)
    assert option1 in option_set
    assert option2 in option_set

def test_same_key_value_option_with_different_key_coexist():
    option_set = OptionSet()
    option1 = KeyValueOption('--o', 'dog', 'rover')
    option2 = KeyValueOption('--o', 'animal', 'rover')

    option_set.put(option1)
    option_set.put(option2)
    assert option1 in option_set
    assert option2 in option_set

def test_all_option_types_coexist():
    option_set = OptionSet()
    option1 = Option('--o')
    option2 = ValueOption('--o', 'animal')
    option3 = KeyValueOption('--o', 'animal', 'rover')

    option_set.put(option1)
    option_set.put(option2)
    option_set.put(option3)
    assert option1 in option_set
    assert option2 in option_set
    assert option3 in option_set

def test_exclusive_options():
    option_set = OptionSet()
    option1 = Option('--o')
    option2 = ValueOption('--p', 'animal')
    option3 = KeyValueOption('--q', 'animal', 'rover')

    option_set.add_mutually_exclusive_options(set([option1, option2, option3]))

    option_set.put(option2)
    with pytest.raises(ValueError):
        option_set.put(option1)






import json
import os
from pathlib import PurePath
from collections import namedtuple
from typing import Optional, Dict

import click
from datetime import datetime

from docker import APIClient
from pytz import timezone


eastern = timezone('US/Eastern')
NETWORK_NAME = os.environ.get('NETWORK_NAME')
USER = os.environ.get('USER')
UID = int(os.environ.get('UID'))
GID = int(os.environ.get('GID'))
DOCKER_GID = int(os.environ.get('DOCKER_GID'))
TRACING_HOST = os.environ.get('TRACING_HOST')
TRACING_PORT = os.environ.get('TRACING_PORT')

# The kernel image where this will be used is supplied by the user so
# we need to use /tmp for the HOME path because it will be writable
home_path_parts = list(PurePath(os.environ.get('HOME')).parts)
HOME = str(PurePath('/tmp').joinpath(*home_path_parts[2:]))


KernelConnection = namedtuple('KernelConnection', ['key', 'control_port', 'shell_port', 'stdin_port', 'hb_port', 'iopub_port'])


__all__ = ['lab_start_kernel', 'lab_start_kernel_impl']


def parse_kernel_connection(connection: str) -> KernelConnection:
    with open(connection, 'r') as cxn_fp:
        connection_params = json.load(cxn_fp)

    return KernelConnection(
        key=connection_params['key'].encode('utf-8'),
        control_port=connection_params['control_port'],
        shell_port=connection_params['shell_port'],
        stdin_port=connection_params['stdin_port'],
        hb_port=connection_params['hb_port'],
        iopub_port=connection_params['iopub_port']
    )


class DockerRunOptions(object):
    def __init__(self):
        self.options = set()

    def with_option(self, option: str) -> 'DockerRunOptions':
        self.options.add(option)
        return self

    def with_http(self, hostname: str, port: int) -> 'DockerRunOptions':
        self.options.add('-l traefik.enable=true')
        self.options.add(f'-l traefik.http.routers.{hostname}.rule=Host(`{hostname}`)')
        self.options.add(f'-l traefik.http.services.{hostname}.loadbalancer.server.port={port}')
        return self

    def with_gpu(self) -> 'DockerRunOptions':
        self.options.add('--gpus all')
        self.options.add('-e NVIDIA_DRIVER_CAPABILITIES=video,compute,utility')
        return self

    def with_privileged(self) -> 'DockerRunOptions':
        self.options.add('--privileged')
        return self

    def with_add_devices(self) -> 'DockerRunOptions':
        self.options.add('-v /dev:/dev')
        self.with_privileged()
        return self

    def with_display(self) -> 'DockerRunOptions':
        display = os.environ.get('DISPLAY')
        self.options.add(f'-e DISPLAY={display}')
        self.options.add('-e QT_X11_NO_MITSHM=1')
        self.options.add('-v /tmp/.X11-unix:/tmp/.X11-unix:ro')
        return self

    def with_shared_memory(self) -> 'DockerRunOptions':
        self.options.add(f'--ipc=host')
        self.options.add('--ulimit memlock=-1:-1')
        self.options.add('--ulimit stack=-1:-1')
        self.with_add_devices()
        return self

    def build(self):
        return ' '.join(self.options)


def lab_start_kernel(connection: str, registry_name: str, options: DockerRunOptions):
    pass


def lab_start_kernel_impl(connection: str, registry_name: str, project_name: str,
                          project_version: str, project_path: str, options: DockerRunOptions):
    options_str = options.build()
    cxn = parse_kernel_connection(connection)
    project_data = os.path.join(project_path, 'data')
    kernel_image = f"{registry_name}/{project_name}:{project_version}"
    cmd = f'docker run --rm --init \
            {options_str} \
            -e USER={USER} \
            -e HOME={HOME} \
            -e UID={UID} \
            -e GID={GID} \
            -e OTEL_EXPORTER_JAEGER_AGENT_HOST={TRACING_HOST} \
            -e OTEL_EXPORTER_JAEGER_AGENT_PORT={TRACING_PORT} \
            -e PROJECT_WORKSPACE={project_path} \
            -e PROJECT_DATA={project_data} \
            -v {project_path}:{project_path} \
            -v {project_data}:{project_data} \
            -w {project_path} \
            --user {UID}:{GID} \
            --network=container:jupyter \
            --name {project_name}-kernel-{ datetime.now(eastern).strftime("%Y-%m-%d__%H-%M-%S") } \
            {kernel_image} \
            launch-kernel \
                --session-key={cxn.key} \
                --heartbeat-port={cxn.hb_port} \
                --shell-port={cxn.shell_port} \
                --iopub-port={cxn.iopub_port} \
                --stdin-port={cxn.stdin_port} \
                --control-port={cxn.control_port}'.split()
    click.echo(f'Starting kernel: {cmd}')
    os.execvp(cmd[0], cmd)


def lab_build_image(registry_name: str, nocache: bool = False, build_args: Optional[Dict[str, str]] = None):
    pass


def lab_build_image_impl(registry_name: str, project_name: str, project_version: str,
                         project_path: str, nocache: bool = False, build_args: Optional[Dict[str, str]] = None):
    image_name = f"{registry_name}/{project_name}:{project_version}"
    client = APIClient(base_url='unix://var/run/docker.sock')
    logs = client.build(
        decode=True,
        nocache=nocache,
        path=project_path,
        tag=image_name,
        rm=True,
        buildargs=build_args
    )
    for chunk in logs:
        if 'stream' in chunk:
            for line in chunk['stream'].splitlines():
                click.echo(line)


def lab_run_image(registry_name: str, project_name: str,
                  project_version: str, project_path: str, options: DockerRunOptions):
    pass


def lab_run_image_impl(registry_name: str, project_name: str,
                       project_version: str, project_path: str, options: DockerRunOptions):
    options_str = options.build()
    project_data = os.path.join(project_path, 'data')
    kernel_image = f"{registry_name}/{project_name}:{project_version}"
    cmd = f'docker run --rm --init \
            {options_str} \
            -e USER={USER} \
            -e HOME={HOME} \
            -e UID={UID} \
            -e GID={GID} \
            -e OTEL_EXPORTER_JAEGER_AGENT_HOST={TRACING_HOST} \
            -e OTEL_EXPORTER_JAEGER_AGENT_PORT={TRACING_PORT} \
            -e PROJECT_WORKSPACE={project_path} \
            -e PROJECT_DATA={project_data} \
            -v {project_path}:{project_path} \
            -v {project_data}:{project_data} \
            -w {project_path} \
            --user {UID}:{GID} \
            --network=container:jupyter \
            --name {project_name}-kernel-{datetime.now(eastern).strftime("%Y-%m-%d__%H-%M-%S")} \
            {kernel_image} \
            launch-kernel \
                --session-key={cxn.key} \
                --heartbeat-port={cxn.hb_port} \
                --shell-port={cxn.shell_port} \
                --iopub-port={cxn.iopub_port} \
                --stdin-port={cxn.stdin_port} \
                --control-port={cxn.control_port}'.split()
    click.echo(f'Running app: {cmd}')
    os.execvp(cmd[0], cmd)

    # "OTEL_EXPORTER_JAEGER_ENDPOINT": "",
    # "OTEL_EXPORTER_JAEGER_AGENT_HOST": "",
    # "OTEL_EXPORTER_JAEGER_AGENT_PORT": "",
    # "OTEL_EXPORTER_JAEGER_AGENT_SPLIT_OVERSIZED_BATCHES": "",
    # "OTEL_EXPORTER_JAEGER_TIMEOUT": "",




from unittest.mock import patch
from typing import Any, Dict, List, Optional

from lab_partner.rootless import LAB_ROOTLESS_DOCKERD, RootlessContainer


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_query_rootless_state(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'name': 'bob'}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert rootless._query_rootless_state(lambda c: 'bob' == c['name'])


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_query_rootless_state_missing(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'name': 'other'}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert not rootless._query_rootless_state(lambda c: 'bob' == c['name'])


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_query_rootless_state_empty(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = []
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert not rootless._query_rootless_state(lambda c: 'bob' == c['name'])


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_query_rootless_state_null(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = None
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert not rootless._query_rootless_state(lambda c: 'bob' == c['name'])


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_does_rootless_container_exist(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'Names': LAB_ROOTLESS_DOCKERD}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert rootless.does_rootless_container_exist()


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_does_rootless_container_not_exist(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'Names': 'bob'}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert not rootless.does_rootless_container_exist()


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_is_rootless_container_running(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'Names': LAB_ROOTLESS_DOCKERD, 'State': 'running'}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert rootless.is_rootless_container_running()


@patch('lab_partner.docker.daemon_info.DockerDaemonInfo')
def test_is_rootless_container_not_running(mock_docker_daemo_info):
    mock_docker_daemo_info.containers.return_value = [{'Names': LAB_ROOTLESS_DOCKERD, 'State': 'stopped'}]
    rootless = RootlessContainer(mock_docker_daemo_info)
    assert rootless.is_rootless_container_not_running()




import os
import pathlib
import logging
import sys
import time
import itertools
from subprocess import CalledProcessError

from . import user_info
from .docker import (
    DockerClient,
    DockerComposeClient,
    DockerEnvironment,
    FilterSet,
    LabelFilter,
    StatusFilter,
    DockerContainerStatus
)
from .process_utils import run_process_quiet
from . import config


logger = logging.getLogger(__name__)


LAB_ROOTLESS_INIT = 'lab-rootless-init'
LAB_ROOTLESS_DOCKERD = 'lab-rootless-dockerd'
LAB_ROOTLESS_STORAGE = 'lab-rootless-storage'
LAB_ROOTLESS_USER_STORAGE = 'lab-rootless-user-storage'
LAB_ROOTLESS_COMPOSE_CONFIG_PATH = 'lab_partner.compose_config.rootless'
LAB_ROOTLESS_COMPOSE_INIT_FILENAME = 'docker-compose-init.yml'
LAB_ROOTLESS_COMPOSE_ROOTLESS_FILENAME = 'docker-compose-rootless.yml'

WAITING_ANIMATION_FRAMES = itertools.cycle(['|', '/', '—', '\\'])


class RootlessContainer:
    def __init__(self, env: DockerEnvironment):
        self.container_name = LAB_ROOTLESS_DOCKERD
        self._docker_client = DockerClient(env)
        self._rootless_docker_client = DockerClient(self.rootless_env())
        self._compose_client = DockerComposeClient(config.project_name(), LAB_ROOTLESS_COMPOSE_CONFIG_PATH, env)
        self._role_label_filter = LabelFilter(config.lab_role_label(), config.lab_container_runtime_role())

    def does_rootless_container_exist(self) -> bool:
        """Test if the rootless container exists in any state on the daemon

        :return: True if a container is found with expected name
        """
        return self._docker_client.has_containers(FilterSet().add_filter(self._role_label_filter))

    def is_rootless_container_running(self) -> bool:
        """Test if the rootless container is running on the daemon

        :return: True if the container is found with the expected name and in the "running" state
        """
        filters = FilterSet().add_filter(self._role_label_filter).add_filter(StatusFilter(DockerContainerStatus.RUNNING))
        return self._docker_client.has_containers(filters)

    def start(self) -> None:
        """Run the sequence of steps to start a rootless docker daemon

        1. Check if rootless is already running. If so then return it.
        2. Create a named volume to store CI/CD artifacts for Act
        3. Finally start a new rootless container

        :param workspace_path: path of the host workspace to bind mount
        :return: DockerDaemonInfo of newly started rootless daemon
        """
        if self.is_rootless_container_running():
            logger.debug('Rootless container already running')
            return

        logger.info('Starting rootless container. It can take a couple minutes to start')
        self._init_rootless()
        self._start_rootless()
        self._rootless_docker_client.create_network(config.lab_network_name())

    def _init_rootless(self) -> None:
        """Create a user accessible named volumed

        Named volumes get create as root. This changes the ownership
        to be "rootless" in order for it to properly get mapped to
        the host user

        :param name: name of the named volume to create
        """
        logger.debug('Intializing rootless environment')
        self._compose_client.run_service(
            LAB_ROOTLESS_INIT,
            LAB_ROOTLESS_COMPOSE_INIT_FILENAME,
            env={
                'LAB_CICD_ARTIFACT_STORAGE': config.cicd_storage_path()
            }
        )

    def _start_rootless(self) -> None:
        """Start rootless container and wait for it to be ready

        :param workspace_path: path of the host workspace to bind mount
        :return: DockerDaemonInfo of newly started rootless daemon
        """
        logger.debug('Starting rootless')
        pathlib.Path(config.cicd_storage_path_host_path()).mkdir(parents=True, exist_ok=True)
        self._compose_client.start_service(
            LAB_ROOTLESS_DOCKERD,
            LAB_ROOTLESS_COMPOSE_ROOTLESS_FILENAME,
            env={
                'HOME': user_info.home(),
                'USER': user_info.username(),
                # 'UID': str(self._user_info.uid()),
                # 'GID': str(self._user_info.gid()),
                'LAB_VERSION': config.version(),
                'LAB_WORKSPACE': config.workspace(),
                'LAB_CICD_ARTIFACT_STORAGE': config.cicd_storage_path(),
                'XDG_RUNTIME_DIR': user_info.user_runtime_tmp_path(),
                'LAB_ROLE_LABEL': config.lab_role_label(),
                'LAB_CONTAINER_RUNTIME_ROLE': config.lab_container_runtime_role()
            }
        )
        # run_rootless_docker_cmd = DockerRunBuilder(self._rootless_docker_image())
        # run_rootless_docker_cmd.options() \
        #     .with_name(self.container_name) \
        #     .with_hostname(self.container_name) \
        #     .with_privileged() \
        #     .with_host_ipc() \
        #     .with_daemon() \
        #     .with_current_user() \
        #     .with_env('POOL_BASE', '172.28.0.0/16') \
        #     .with_env('POOL_SIZE', '21') \
        #     .with_port_mapping(80, 80) \
        #     .with_bind_mount(workspace_path, workspace_path) \
        #     .with_bind_mount('/tmp', '/tmp') \
        #     .with_bind_mount('/dev', '/dev') \
        #     .with_named_volume(LAB_ROOTLESS_STORAGE, '/var/lib/docker') \
        #     .with_named_volume(LAB_ROOTLESS_USER_STORAGE, '/home/rootless/.local/share/docker') \
        #     .with_named_volume(LAB_CICD_ARTIFACT_STORAGE, '/opt/lab/cicd/artifacts') \
        #     .with_mount_home() \
        #     .with_mount_user_run()
        # rendered_cmd = run_rootless_docker_cmd.build()
        # logger.debug(rendered_cmd)
        # for log_line in run_process(rendered_cmd):
        #     logger.info(log_line)
        self._wait_for_rootless()

    def _wait_for_rootless(self, timeout: int = 120) -> None:
        """Waits for rootless container to start accepting API calls

        :param timeout: number of seconds to wait for, defaults to 120
        :raises TimeoutError: timeout error raised when exeeds timeout
        """
        start_time = time.perf_counter()
        while True:
            if not self.is_rootless_container_running():
                time.sleep(1)
                continue
            try:
                run_process_quiet('docker info', {'DOCKER_HOST': self._rootless_docker_url()})
                break
            except CalledProcessError as ex:
                elapsed_time = time.perf_counter() - start_time
                sys.stdout.write(f'{next(WAITING_ANIMATION_FRAMES)} Waiting for rootless docker to start\r')
                sys.stdout.flush()
                if elapsed_time >= timeout:
                    raise TimeoutError(f'Timeout waiting for rootless docker after {timeout} seconds', ex)
                time.sleep(.1)

    def stop(self) -> None:
        if not self.is_rootless_container_running():
            logger.debug('Rootless container is not running')
            return

        logger.debug('About to stop rootless')
        pathlib.Path(config.cicd_storage_path_host_path()).mkdir(parents=True, exist_ok=True)
        self._compose_client.stop_service(
            LAB_ROOTLESS_DOCKERD,
            LAB_ROOTLESS_COMPOSE_ROOTLESS_FILENAME,
            env={
                'HOME': user_info.home(),
                'USER': user_info.username(),
                # 'UID': str(self._user_info.uid()),
                # 'GID': str(self._user_info.gid()),
                'LAB_VERSION': config.version(),
                'LAB_WORKSPACE': config.workspace(),
                'LAB_CICD_ARTIFACT_STORAGE': config.cicd_storage_path(),
                'XDG_RUNTIME_DIR': user_info.user_runtime_tmp_path(),
                'LAB_ROLE_LABEL': config.lab_role_label(),
                'LAB_CONTAINER_RUNTIME_ROLE': config.lab_container_runtime_role()
            }
        )

    @staticmethod
    def _rootless_docker_image() -> str:
        """Build the docker images name based on the python
        package version

        :return: name of the rootless docker image to use
        """
        return f'enclarify/lab-partner-dind-rootless:{config.version()}'

    @staticmethod
    def _rootless_docker_url() -> str:
        """Construct the rootless docker file socket url
        based on XDG_RUNTIME_DIR environment variable. Likely needs to be
        fixed for other platforms like mac

        :return: rootless file socket url
        """
        xdg_runtime_dir = os.environ['XDG_RUNTIME_DIR']
        return f'unix://{xdg_runtime_dir}/docker.sock'

    @classmethod
    def rootless_env(cls) -> DockerEnvironment:
        return DockerEnvironment.build_with_overridden_host(cls._rootless_docker_url())



import sys

import click
from lab_partner_utils.cli_command_factory import WorkspaceEntryPointMetadata
from lab_partner_utils.utils import start_services, stop_services, read_env


if sys.version_info >= (3, 12):
    from importlib import resources
else:
    import importlib_resources as resources


data_text = resources.files('lab_partner_tools.data').joinpath('data1.txt').read_text()


WORKSPACE_METADATA_SHIM = WorkspaceEntryPointMetadata('', '', '', '', 'lab', cli_root, os.path.join(cli_root, 'data'))
WEB_SERVICE = 'proxy'



@click.group()
def web_proxy():
    pass


@web_proxy.command()
@click.option('--listen-on-high-port', is_flag=True, show_default=True, default=False, help='Listen on port 8600 instead of 80. Useful if sitting behind another proxy already')
def start_web(listen_on_high_port: bool) -> None:
    """
    Start the reverse proxy webserver
    """
    env_override = {}
    if listen_on_high_port:
        env_override['TRAEFIK_LISTEN_PORT'] = '8600'
    start_services(WORKSPACE_METADATA_SHIM, ENVIRONMENT_PATH, services=WEB_SERVICE, env_override=env_override)


if __name__ == '__main__':
    web_proxy()



#!/usr/bin/env python

import os
import pathlib

from ipykernel import kernelapp as app
import click


@click.command()
@click.option('--session-key', required=True)
@click.option('--heartbeat-port', required=True, type=int)
@click.option('--shell-port', required=True, type=int)
@click.option('--iopub-port', required=True, type=int)
@click.option('--stdin-port', required=True, type=int)
@click.option('--control-port', required=True, type=int)
@click.option('--debug', default=True)
def launch(session_key, heartbeat_port, shell_port, iopub_port, stdin_port, control_port, debug: bool):
    home = os.environ.get('HOME')
    uid = int(os.environ.get('UID'))
    gid = int(os.environ.get('GID'))
    pathlib.Path(home).mkdir(parents=True, exist_ok=True)
    os.chown(home, uid, gid)
    args = [
        f'--Session.key={session_key}',
        f'--ip=0.0.0.0',
        f'--hb={heartbeat_port}',
        f'--shell={shell_port}',
        f'--iopub={iopub_port}',
        f'--stdin={stdin_port}',
        f'--control={control_port}'
    ]
    if debug:
        args.append('--debug')
    app.launch_new_instance(argv=args)


if __name__ == '__main__':
    launch()



import os
from typing import Dict, Optional, List, Callable, Any
import sys
import logging

import click
from click import Command, Context, MultiCommand, CommandCollection
from ..utils import WorkspaceEntryPointMetadata
from . import (
    CLI_PLUGINS_ENTRY_POINT,
    CLI_WORKSPACE_ONLY_PLUGINS_ENTRY_POINT
)
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


WORKSPACE_PATH = os.environ['LAB_WORKSPACE']
WORKSPACE_DATA_PATH = os.path.join(WORKSPACE_PATH, 'data')


logging.basicConfig(level='DEBUG')
logger = logging.getLogger(__name__)


class WorkspaceEntrypointCommands(MultiCommand):
    def list_commands(self, ctx: Context) -> List[str]:
        """
        Generates list of subcommand names by scanning the workspace for cli_plugins entry point

        :param ctx: current click context
        :return: list of subcommand names
        """
        click.echo('here')
        raise ValueError('list commands')
        found_cli_plugins = self._cli_plugin_search(WORKSPACE_PATH)
        plugin_names = []
        if found_cli_plugins:
            plugin_names = list(found_cli_plugins.keys())
            plugin_names.sort()
        return plugin_names

    def get_command(self, ctx: Context, project_name: str) -> Optional[Command]:
        """
        Builds a set of commands for a specific project and returns a click Command

        :param ctx: current click context
        :param project_name: name of the scanned project with CLI an entry point
        :return: Compiled click Command or None if there was an error
        """
        click.echo('here')
        raise ValueError('list commands')
        found_cli_plugins = self._cli_plugin_search(WORKSPACE_PATH)
        if project_name not in found_cli_plugins:
            return None

        plugin_metadata = found_cli_plugins[project_name]
        plugin_script_path = self._absolute_script_path(plugin_metadata.project_path, plugin_metadata.script_name)
        if not plugin_script_path:
            logger.error(
                f'Error: project "{project_name}" supplies a CLI plugin but script "{plugin_metadata.script_name}" could not be loaded')
            return None
        ctx.meta.update(plugin_metadata.__dict__)
        return self.compile_command(plugin_script_path, plugin_metadata.function_name)

    @staticmethod
    def compile_command(plugin_script_path, function_name) -> Optional[Command]:
        """
        Compiles click Command from the specified script path. This allows discovered workspace
        commands to be used on the fly

        :param plugin_script_path: absolute path to the script file to compile
        :param function_name: name of the function to pull the compiled command out of scope
        :return: Compiled click Command or None if there was an error
        """
        with open(plugin_script_path) as f:
            code = compile(f.read(), plugin_script_path, mode='exec')
            try:
                ns = {}
                eval(code, ns, ns)
                return ns[function_name]
            except Exception:
                logger.exception(f'Error compiling {plugin_script_path}')
                return None

    @classmethod
    def _cli_plugin_search(cls, workspace_path: str) -> Dict[str, WorkspaceEntryPointMetadata]:
        """
        Searches the workspace looking for pyproject.toml files with a cli.cli_plugins
        entry point

        :param workspace_path: path of the workspace to search
        :return: a dictionary of each entry point found
        """
        cli_plugins = {}
        print(f'About to search for plugins')
        for project_name in os.listdir(workspace_path):
            project_path = os.path.join(workspace_path, project_name)
            project_data_path = os.path.join(project_path, 'data')

            project_toml_path = os.path.join(project_path, 'pyproject.toml')
            if not os.path.exists(project_toml_path):
                print(f'Path does not have a pyproject.toml')
                continue

            with open(project_toml_path, "rb") as f:
                project_toml = tomllib.load(f)
                has_entrypoint = cls.check_for_entrypoint(CLI_PLUGINS_ENTRY_POINT, project_toml)
                has_workspace_entrypoint = cls.check_for_entrypoint(CLI_WORKSPACE_ONLY_PLUGINS_ENTRY_POINT, project_toml)
                if not has_entrypoint and not has_workspace_entrypoint:
                    logger.debug(f'the project {project_name} does not have entrypoint')
                    continue

                project_version = "5"
                if has_entrypoint:
                    found_entry_points = project_toml['project']['entry-points'][CLI_PLUGINS_ENTRY_POINT]
                    for entry_point_name, entry_point_value in found_entry_points.items():
                        entry_point_metadata = WorkspaceEntryPointMetadata.build(entry_point_name, entry_point_value, project_name, project_version, project_path, project_data_path)
                        cli_plugins[entry_point_metadata.name] = entry_point_metadata
                if has_workspace_entrypoint:
                    found_entry_points = project_toml['project']['entry-points'][CLI_WORKSPACE_ONLY_PLUGINS_ENTRY_POINT]
                    for entry_point_name, entry_point_value in found_entry_points.items():
                        entry_point_metadata = WorkspaceEntryPointMetadata.build(entry_point_name, entry_point_value, project_name, project_version, project_path, project_data_path)
                        cli_plugins[entry_point_metadata.name] = entry_point_metadata
        return cli_plugins

    @staticmethod
    def check_for_entrypoint(entrypoint_name: str, project_toml: Dict[str, Any]) -> bool:
        return 'project' in project_toml \
            and 'entry-points' in project_toml['project'] \
            and entrypoint_name in project_toml['project']['entry-points']

    @staticmethod
    def _absolute_script_path(project_path: str, script_name: str) -> Optional[str]:
        """
        Walks the filesystem below the project path to resolve the absolute path of
        the script specified in the entry point

        :param project_path: absolute path to the project
        :param script_name: name of the script somewhere in the project path
        :return: absolute path of the project
        """
        script_filename = f'{script_name}.py'
        exclude = {'.git', '.idea', '.ipynb_checkpoints', '__pycache__', '.pytest_cache', 'data'}
        for root, dirs, files in os.walk(project_path):
            dirs[:] = [d for d in dirs if d not in exclude]
            absolute_script_path = os.path.join(root, script_filename)
            if script_filename in files and os.path.exists(absolute_script_path):
                return absolute_script_path





import logging
import sys
import os
import shlex
from subprocess import (
    Popen,
    PIPE,
    STDOUT,
    CalledProcessError,
    check_call,
    DEVNULL
)
import json
from typing import Iterable, Dict, Any, List, Optional

from .types import ShellCommand


logger = logging.getLogger(__name__)


def run_process_quiet(command: ShellCommand, env: Optional[Dict[str, str]] = None):
    command_parts = shlex.split(command) if isinstance(command, str) else command
    check_call(command_parts, stdout=DEVNULL, stderr=DEVNULL, env=env)


def run_process_stream_result(command: ShellCommand, env: Optional[Dict[str, str]] = None) -> Iterable[str]:
    """Run subprocess that redirects stderr to stdout to support applications
    that use stderr to separate logs types rather than just signalling error.

    :param command: the command line string to be executed
    :return: a generator of each returned line from STDOUT
    """
    command_parts = shlex.split(command) if isinstance(command, str) else command
    with Popen(command_parts, stdout=PIPE, stderr=STDOUT, bufsize=1, universal_newlines=True, env=env, encoding='utf-8') as p:
        if p.stdout:
            for line in p.stdout:
                if line:
                    yield line
    if p.returncode != 0:
        raise CalledProcessError(p.returncode, p.args)


def process_output_as_json(process_output: Iterable[str]) -> List[Dict[str, Any]]:
    """Convert the results of a subprocess to a json generator

    :param process_output: output from running `run_process(...)`
    :return:
    """
    return [json.loads(output) for output in process_output]


def run_process(command: ShellCommand, env: Optional[Dict[str, str]] = None) -> List[str]:
    return list(run_process_stream_result(command, env))


def run_process_single_result(command: ShellCommand, env: Optional[Dict[str, str]] = None) -> str:
    return run_process(command, env)[0]


def exec_process(command: ShellCommand, env: Optional[Dict[str, str]] = None) -> None:
    """
    Exec a  process by forking the current process allowing the child
    to take over the main process id

    :param command: any command line string to execute
    :param input: text content to send to STDIN of the process
    """
    command_parts = shlex.split(command) if isinstance(command, str) else command
    if env:
        os.execvpe(command_parts[0], command_parts, env)
    else:
        os.execvp(command_parts[0], command_parts)


def exec_process_with_input(command: str, input: str, env: Optional[Dict[str, str]] = None) -> None:
    """
    Exec a  process by forking the current process allowing the child
    to take over the main process id

    :param command: any command line string to execute
    :param input: text content to send to STDIN of the process
    """
    r, w = os.pipe()
    pid = os.fork()
    if pid == 0:
        # Child process
        os.dup2(r, sys.stdin.fileno())
        os.close(r)
        os.close(w)
        command_parts = shlex.split(command)
        if env:
            os.execvpe(command_parts[0], command_parts, env)
        else:
            os.execvp(command_parts[0], command_parts)
    else:
        # Parent process
        os.close(r)
        os.write(w, bytearray(input, 'utf-8'))
        os.close(w)
        pid, status = os.waitpid(pid, 0)



import os
import sys
from importlib import metadata
import pathlib
main_dir = 'my_main_dir'
sub_dir = 'sub_dir'
fname = 'filename.tsv'


LAB_PROJECT_NAME = 'lab-partner'
LAB_CICD_ARTIFACT_STORAGE = '/opt/lab/cicd/artifacts'
LAB_CICD_ARTIFACT_STORAGE_HOST_PATH = '/tmp/lab-cicd'


def project_name() -> str:
    return LAB_PROJECT_NAME


def version() -> str:
    return metadata.version(LAB_PROJECT_NAME)


def workspace() -> str:
    return os.environ['LAB_WORKSPACE']


def workspace_data() -> str:
    return str(pathlib.Path(workspace(), ''))


def cicd_storage_path() -> str:
    return LAB_CICD_ARTIFACT_STORAGE


def cicd_storage_path_host_path() -> str:
    return LAB_CICD_ARTIFACT_STORAGE_HOST_PATH


def is_linux() -> bool:
    """
    Check current platform is Linux
    :return: True on Linux
    """
    return sys.platform in ('linux',)


def is_supported_platform() -> bool:
    """
    Check current platform is MacOS or Linux
    :return: True on MacOS or Linux
    """
    return sys.platform in ('darwin', 'linux')


def lab_role_label():
    return 'lab-role'


def lab_container_runtime_role():
    return 'container-runtime'


def lab_shell_role():
    return 'shell'


def lab_network_name():
    return 'lab'



from enum import Enum
import json
import logging
from typing import Any, Dict, List, Optional, Set
from importlib.resources import files
from lab_partner.process_utils import (
    process_output_as_json,
    run_process,
    run_process_single_result,
    exec_process,
    run_process_stream_result
)

logger = logging.getLogger(__name__)


class FilterOperator(Enum):
    EQUAL = '='
    NOT_EQUAL = '!='


class DockerContainerStatus(Enum):
    CREATED = 'created'
    RUNNING = 'running'
    PAUSED = 'paused'
    RESTARTING = 'restarting'
    EXITED = 'exited'
    REMOVING = 'removing'
    DEAD = 'dead'


class LabelFilter:
    def __init__(
            self, key: str, value: Optional[str] = None,
            key_operator: FilterOperator = FilterOperator.EQUAL,
            value_operator: FilterOperator = FilterOperator.EQUAL) -> None:
        self._key = key
        self._value = value
        self._key_operator = key_operator
        self._value_operator = value_operator

    def __str__(self) -> str:
        s = 'label' + self._key_operator.value + self._key
        if self._value:
            s += self._value_operator.value + self._value
        return s


class StatusFilter:
    def __init__(self, status: DockerContainerStatus, operator: FilterOperator = FilterOperator.EQUAL) -> None:
        self._status = status
        self._operator = operator

    def __str__(self) -> str:
        return 'status' + self._operator.value + self._status.value


class FilterSet:
    def __init__(self) -> None:
        self._options: Set[StatusFilter | LabelFilter] = set()

    def add_filter(self, filter: StatusFilter | LabelFilter) -> 'FilterSet':
        self._options.add(filter)
        return self

    def __str__(self) -> str:
        return ' '.join(map(lambda x: '-f ' + str(x), self._options))


class DockerEnvironment:
    def __init__(self, initial_env: Optional[Dict[str, str]] = None) -> None:
        self._env = initial_env or {}

    def put(self, env: Optional[Dict[str, str]]) -> 'DockerEnvironment':
        return DockerEnvironment({**self._env, **(env or {})})

    def as_dict(self) -> Dict[str, str]:
        return self._env

    @staticmethod
    def build_with_overridden_host(docker_host: str):
        return DockerEnvironment({'DOCKER_HOST': docker_host})


class DockerComposeClient:
    def __init__(self, project_name: str, config_root_path: str, env: DockerEnvironment) -> None:
        self._project_name = project_name
        self._config_root_path = config_root_path
        self._env = env

    def exec_service(self, service_name: str, config_filename: str = 'docker-compose.yml', env: Optional[Dict[str, str]] = None) -> None:
        config = files(self._config_root_path).joinpath(config_filename)
        cmd = f'docker compose -f {config} -p {self._project_name} run --remove-orphans {service_name}'
        logger.debug(cmd)
        exec_process(cmd, env=self._env.put(env).as_dict())

    def run_service(self, service_name: str, config_filename: str = 'docker-compose.yml', env: Optional[Dict[str, str]] = None) -> None:
        config = files(self._config_root_path).joinpath(config_filename)
        cmd = f'docker compose -f {config} -p {self._project_name} run -T --remove-orphans {service_name}'
        logger.debug(cmd)
        logs = run_process_stream_result(cmd, env=self._env.put(env).as_dict())
        for log in logs:
            logger.info(log)

    def start_service(self, service_name: str, config_filename: str = 'docker-compose.yml', env: Optional[Dict[str, str]] = None) -> None:
        config = files(self._config_root_path).joinpath(config_filename)
        cmd = f'docker compose -f {config} -p {self._project_name} up -d --remove-orphans --force-recreate --no-build {service_name}'
        logger.debug(cmd)
        logs = run_process_stream_result(cmd, env=self._env.put(env).as_dict())
        for log in logs:
            logger.info(log)

    def stop_service(self, service_name: str, config_filename: str = 'docker-compose.yml', env: Optional[Dict[str, str]] = None):
        config = files(self._config_root_path).joinpath(config_filename)
        cmd = f'docker compose -f {config} -p {self._project_name} stop {service_name}'
        logger.debug(cmd)
        logs = run_process_stream_result(cmd, env=self._env.put(env).as_dict())
        for log in logs:
            logger.info(log)


class DockerClient:
    def __init__(self, env: DockerEnvironment):
        self._env = env

    def network_exists(self, network_name: str) -> bool:
        for net in self.read_networks():
            if network_name == net['Name']:
                return True
        return False

    def create_network(self, network_name: str) -> None:
        """
        Creates a bridged network
        :return: None
        """
        if not self.network_exists(network_name):
            logger.info(f'Creating network: {network_name}')
            cmd = f'docker network create {network_name}'
            rs = run_process(cmd, self._env.as_dict())
            for line in rs:
                logger.debug(line)
        else:
            logger.debug(f'Network {network_name} already exists')

    def volume_exists(self, volume_exists: str) -> bool:
        for net in self.read_volumes():
            if volume_exists == net['Name']:
                return True
        return False

    def create_volume(self, name: str) -> bool:
        """Create a named volume if it doesn't already exist

        :param name: name of the volume to create
        :return: was a volume created?
        """
        if not self.volume_exists(name):
            logger.info(f'Creating volume: {name}')
            rs = run_process(f'docker volume create {name}', self._env.as_dict())
            for line in rs:
                logger.debug(line)
            return True
        else:
            logger.debug(f'Volume "{name}" already exists')
            return False

    def read_daemon_info(self) -> Dict[str, Any]:
        """
        Returns the output of `docker info` as a dictionary
        :return:
        """
        info_str = run_process_single_result('docker info --format "{{json .}}"', self._env.as_dict())
        return json.loads(info_str)

    def has_containers(self, filters: Optional[FilterSet] = None) -> bool:
        return True if self.read_containers(filters) else False

    def read_containers(self, filters: Optional[FilterSet] = None) -> List[Dict[str, Any]]:
        filters_str = str(filters) if filters else ''
        cmd = 'docker container ls -a --format "{{json .}}"'
        rs = run_process(' '.join([cmd, filters_str]), self._env.as_dict())
        containers = process_output_as_json(rs)
        logger.debug('Docker containers:')
        logger.debug(containers)
        return containers

    def kill_container(self, container_name: str) -> None:
        for log_line in run_process(f'docker rm -f {container_name}'):
            logger.info(log_line)

    def read_networks(self) -> List[Dict[str, Any]]:
        rs = run_process('docker network ls --format "{{json .}}"', self._env.as_dict())
        nets = process_output_as_json(rs)
        logger.debug('Docker networks:')
        logger.debug(nets)
        return nets

    def read_volumes(self) -> List[Dict[str, Any]]:
        rs = run_process('docker volume ls --format "{{json .}}"', self._env.as_dict())
        vols = process_output_as_json(rs)
        logger.debug('Docker volumes:')
        logger.debug(vols)
        return vols



import logging
from pathlib import Path
import os
os.environ['LAB_WORKSPACE'] = f'{str(Path(__file__).parent.resolve())}/fixtures/workspace_commands'

from click.testing import CliRunner
from lab_partner_utils.command_sets.workspace_entrypoint_commands import WorkspaceEntrypointCommands




def test_commands():
    print(f'{str(Path(__file__).parent.resolve())}/fixtures/workspace_commands')
    runner = CliRunner()
    result = runner.invoke(WorkspaceEntrypointCommands, ['hello'])
    print(result.stdout)
    assert result.exit_code == 0
    # print(result.stderr)


from lab_partner.docker import (
    DockerContainerStatus, 
    FilterOperator, 
    FilterSet, 
    LabelFilter, 
    StatusFilter
)


def test_status_filter():
    filter = StatusFilter(DockerContainerStatus.RUNNING)
    assert str(filter) == 'status=running'


def test_status_filter_not():
    filter = StatusFilter(DockerContainerStatus.RUNNING, FilterOperator.NOT_EQUAL)
    assert str(filter) == 'status!=running'


def test_label_filter():
    filter = LabelFilter('name')
    assert str(filter) == 'label=name'


def test_label_filter_not():
    filter = LabelFilter('name', key_operator=FilterOperator.NOT_EQUAL)
    assert str(filter) == 'label!=name'


def test_label_filter_with_value():
    filter = LabelFilter('name', 'bob')
    assert str(filter) == 'label=name=bob'


def test_label_filter_not_with_value():
    filter = LabelFilter('name', 'bob', key_operator=FilterOperator.NOT_EQUAL, value_operator=FilterOperator.NOT_EQUAL)
    assert str(filter) == 'label!=name!=bob'


def test_filterset():
    status_filter = StatusFilter(DockerContainerStatus.RUNNING)
    label_filter = LabelFilter('name')
    all_filters = FilterSet() \
        .add_filter(status_filter) \
        .add_filter(label_filter) \
        .add_filter(label_filter)
    assert str(all_filters) == '-f label=name -f status=running'


import typing as t
import inspect
import importlib

import click
from click import Context, Command, Option, Argument
import docstring_parser


class CommandBuilder(click.MultiCommand):

    def __init__(self, fn_list: t.List[t.Callable], name: t.Optional[str] = None, invoke_without_command: bool = False,
                 no_args_is_help: t.Optional[bool] = None, subcommand_metavar: t.Optional[str] = None,
                 chain: bool = False, result_callback: t.Optional[t.Callable[..., t.Any]] = None,
                 **attrs: t.Any) -> None:
        super().__init__(name, invoke_without_command, no_args_is_help, subcommand_metavar, chain, result_callback, **attrs)
        self.fn_map = {fn.__name__: inspect.getmodule(fn).__name__ for fn in fn_list}

    def list_commands(self, ctx: Context) -> t.List[str]:
        command_names = list(self.fn_map.keys())
        command_names.sort()
        return command_names

    def get_command(self, ctx: Context, cmd_name: str) -> t.Optional[Command]:
        module = importlib.import_module(self.fn_map[cmd_name])
        fn = getattr(module, cmd_name)

        docstring = docstring_parser.parse(inspect.getdoc(fn))
        help = f'''
        {docstring.short_description}
        
        {docstring.long_description}
        '''
        short_help = docstring.short_description
        fn_sig = inspect.signature(fn)

        options = []
        for i, param in enumerate(fn_sig.parameters.values()):
            arg_type = click.types.convert_type(param.annotation)
            arg_name = param.name.replace('_', '-')
            print(f'param.annotation: {param.annotation}')
            arg_description = docstring.params[i].description







import os
import logging
from pathlib import Path
from typing import List, Optional, Set

from lab_partner.option_set import Option, OptionSet
from ..user_info import UnixUser


logger = logging.getLogger(__name__)


class InvalidDockerOptionConfiguration(Exception):
    pass


def environment_variable_renderer(key: str, value: Optional[str] = None) -> str:
    if value:
        return f'{key}={value}'
    else:
        return key

def named_volume_mount_renderer(key: str, value: str):
    return f'type=volume,src={key},dst={value}'

def bind_mount_render(key: str, value: str):
    return f'type=bind,src={key},dst={value}'


class DockerRunOptions:
    """
    This class helps build up a list of options for the invocation of Docker.
    """
    def __init__(self):
        self._options = OptionSet()
        self._tty_option = Option('-it')
        self._daemon_option = Option('-d')
        self._options.add_mutually_exclusive_options((self._tty_option, self._daemon_option))
        self._user_info = UnixUser()

    def with_init(self) -> 'DockerRunOptions':
        """Add the '--init' option to run Tini to manage the
        main process

        :return: this DockerRunOptions instance
        """
        self._options.put_option('--init')
        return self

    def with_remove_on_exit(self) -> 'DockerRunOptions':
        """Add the '--rm' option to remove the container 
        after exiting

        :return: this DockerRunOptions instance
        """
        self._options.put_option(f'--rm')
        return self

    def with_name(self, name: str) -> 'DockerRunOptions':
        """Add the '--name' option to name the container

        :param name: name of the container
        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--name', name)
        return self

    def with_hostname(self, hostname: str) -> 'DockerRunOptions':
        """Add the '--hostname' option to set the container
        hostname

        :param hostname: hostname of the container
        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--hostname', hostname)
        return self
    
    def with_user(self, uid: int, gid: int) -> 'DockerRunOptions':
        """Add the '--user' option to run the container 
        as a specific user and group

        :param uid: UID of the user to run as
        :param gid: GID of the user to run as
        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--user', f'{uid}:{gid}')
        return self

    def with_current_user(self) -> 'DockerRunOptions':
        """Add the '--user' option with the current
        user and group

        :return: this DockerRunOptions instance
        """
        return self.with_user(self._user_info.uid, self._user_info.gid)
    
    def with_root_user(self) -> 'DockerRunOptions':
        """Add the '--user' option with the 0 as the UID and GID to force
        running as root

        :return: this DockerRunOptions instance
        """
        return self.with_user(0, 0)

    def with_privileged(self) -> 'DockerRunOptions':
        """Add the '--privileged' flag to run the container with no limitations

        :return: this DockerRunOptions instance
        """
        self._options.put_option('--privileged')
        return self
    
    def with_host_ipc(self) -> 'DockerRunOptions':
        """Add the '--ipc' option set to 'host' to share all IPC resources

        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--ipc', 'host')
        return self

    def with_tty(self) -> 'DockerRunOptions':
        """Add the '-it' option to start interactive TTY

        :return: this DockerRunOptions instance
        """
        self._options.put(self._tty_option)
        return self

    def with_daemon(self) -> 'DockerRunOptions':
        """Add the '-d' option to start in daemon mode

        :return: this DockerRunOptions instance
        """
        self._options.put(self._daemon_option)
        return self

    def with_network(self, network_name: str) -> 'DockerRunOptions':
        """Add the '--network' option to specify the network name

        :param network_name: name of the network
        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--network', network_name)
        return self

    def with_workdir(self, workdir: str) -> 'DockerRunOptions':
        """Add the '--workdir' option to specify the current working dirtory 

        :param workdir: path to current working directory
        :return: this DockerRunOptions instance
        """
        self._options.put_value_option('--workdir', workdir)
        return self

    def with_env(self, key: str, value: Optional[str] = None) -> 'DockerRunOptions':
        """Add the '-e' environment value option to set environment key/value pairs. However,
        docker does allow environment variables to set without values if they are currently set
        as a shortcut. 

        :param key: Environment variable name
        :param value: Environment variable value
        :return: this DockerRunOptions instance
        """
        self._options.put_key_value_option('-e', key, value, environment_variable_renderer)
        return self

    def with_mount_user_run(self) -> 'DockerRunOptions':
        """Add the '--mount' option as a bind mount of the user run path of the 
        system. This is the same as XDG_RUNTIME_DIR on a linux system.

        :return: this DockerRunOptions instance
        """
        self._options.put_key_value_option('--mount', f'/run/user/{self._user_info.uid}/', f'/run/user/{self._user_info.uid}/', bind_mount_render)
        return self

    def with_mount_home(self) -> 'DockerRunOptions':
        """Add the '--mount' option as a bind mount of the user runtime path of the 
        system. This is the same as XDG_RUNTIME_DIR on a linux system.

        :return: this DockerRunOptions instance
        """
        self._options.put_key_value_option('--mount', self._user_info.home, self._user_info.home, bind_mount_render)
        return self

    def with_home_dir_bind_mount(self, source: str, target: str, validate_source_exists: bool = True) -> 'DockerRunOptions':
        source_in_home = self._user_info.home_subdir(source)
        if source and target and self._path_exists(source_in_home, validate_source_exists):
            self._options.put_key_value_option('--mount', source_in_home, target, bind_mount_render)
        else:
            logger.warning(f'Requested HOME mount that does not exist: {source_in_home}')
        return self

    def with_bind_mount(self, source: str, target: str) -> 'DockerRunOptions':
        """Adds an option to bind mount a host volume

        :param source: Source host path to be mounted
        :param target: Target path inside container to attach the mount
        :return: self
        """
        self._options.add(f'-v {source}:{target}')
        return self

    def with_named_volume(self, name: str, target: str) -> 'DockerRunOptions':
        self._options.add(f'--mount type=volume,src={name},dst={target}')
        return self

    def with_port_mapping(self, external_port: int, internal_port: int) -> 'DockerRunOptions':
        self._options.add(f'-p {external_port}:{internal_port}')
        return self
    
    def with_entrypoint(self, entrypoint: str) -> 'DockerRunOptions':
        self._options.add(f'--entrypoint={entrypoint}')
        return self

    def build(self) -> str:
        """Builds the accumulated options into a space-separated string.

        :return: String containing all the options.
        """
        return str(self._options)

    @staticmethod
    def _path_exists(path: str, should_validate: bool) -> bool:
        """Check for the existence of a source path

        :param path: Path to be checked
        :param should_validate: Whether we should really check.  If False, the method
        will return True regardless of whether the path exists.
        :return:
        """
        if not should_validate:
            return True
        else:
            return Path(path).exists()
    

class DockerRunBuilder(object):
    def __init__(self, image_name: str, command: str = ''):
        self._image_name = image_name
        self._command = command
        self._options = DockerRunOptions()

    def options(self):
        return self._options

    def build(self) -> str:
        return f'docker run \
                {self._options.build()} \
                {self._image_name} \
                {self._command}'



import logging
import os
import sys
from typing import Dict, Tuple, Optional
import subprocess
import click
from click import Context
from dotenv import dotenv_values
from docker import APIClient


logger = logging.getLogger(__name__)


class WorkspaceEntryPointMetadata(object):
    """
    Scans the workspace looking for specific entry points able to be
    parsed for CLI commands to dynamically add
    """

    def __init__(self, name: str, value: str, script_name: str, function_name: str,
                 project_name: str, project_version: str, project_path: str, project_data_path: Optional[str] = None):
        """
        Constructor

        :param name: parsed name of the cli_plugins entry point
        :param value: parsed value of the cli_plugins entry point
        :param script_name: parsed script name of the cli_plugins entry point
        :param function_name: parsed function name of the cli_plugins entry point
        :param project_name: name of the project based on the directory the pyproject.toml is found in
        :param project_version: version as found in the pyproject.toml
        :param project_path: path of the project in the workspace containing a cli_plugins
        :param project_data_path: path to an optional 'data' folder in the project
        """
        self.name = name
        self.value = value
        self.script_name = script_name
        self.function_name = function_name
        self.project_name = project_name
        self.project_version = project_version
        self.project_path = project_path
        self.project_data_path = project_data_path

    @staticmethod
    def from_context(ctx: Context) -> 'WorkspaceEntrypointMetadata':
        return WorkspaceEntryPointMetadata(
            ctx.meta['name'],
            ctx.meta['value'],
            ctx.meta['script_name'],
            ctx.meta['function_name'],
            ctx.meta['project_name'],
            ctx.meta['project_version'],
            ctx.meta['project_path'],
            ctx.meta['project_data_path'])

    @classmethod
    def build(cls, entry_point_name: str, entry_point_value: str, project_name: str, project_version: str, project_path: str, project_data_path: str) -> 'WorkspaceEntrypointMetadata':
        """
        Static builder. Parses a raw entry point string and builds a WorkspaceEntrypointMetadata

        :param entry_point_name: name of the entry-point in a pyproject.toml file
        :param entry_point_value: value of the entry-point in a pyproject.toml file
        :param project_name: name of the directory the pyproject.toml is found in
        :param project_version: version as found in the pyproject.toml
        :param project_path: path of the project in the workspace containing a cli_plugins
        :param project_data_path: path to an optional 'data' folder in the project
        :return:
        """
        cls._validate_entry_point_value(entry_point_name, entry_point_value)
        script_name, function_name = cls._parse_entry_point_value(entry_point_value)
        return WorkspaceEntryPointMetadata(entry_point_name, entry_point_value, script_name, function_name,
                                           project_name, project_version, project_path, project_data_path)

    @staticmethod
    def _validate_entry_point_value(entry_point_name: str, entry_point_value: str) -> None:
        """
        Validates a raw entry point string to ensure it is well-formed and
        raises a ValueError if not

        Should be of the form:
            <command-name> = <folder_name>.<script_name>:<function_name>

        :param entry_point_value: raw entry point string in a pyproject.toml file
        """
        # TODO make if condition use a proper regex for the format below
        if ':' not in entry_point_value:
            error_message = f'''
            !----------------------------------------------------
            Found CLI plugin entry point of:
                {entry_point_name} = "{entry_point_value}"

            CLI plugin entry point not well formed. Should be of the form:
                <command-name> = "<folder_name>.<script_name>:<function_name>"

            For example: 
                example = "example_project_command_set:example"

            The <folder_name> is optional
            !---------------------------------------------------- 
            '''
            raise ValueError(error_message)

    @staticmethod
    def _parse_entry_point_value(entry_point_value: str) -> Tuple[str, str]:
        """
        Parse entry point value string into a tuple of:
            script name
            function name

        :param entry_point_value: entry point value string in a pyproject.toml file
        :return:
        """
        script_path, function_name = entry_point_value.split(':')
        script_path_parts = script_path.split('.')
        script_name = script_path_parts[-1]
        return script_name, function_name


def print_subprocess_output(rs: subprocess.CompletedProcess) -> None:
    """
    Prints stdout of subprocess

    :param rs: a CompletedProcess to print stdout from
    """
    if rs.returncode == 0:
        click.echo(rs.stdout.decode('utf-8'))
    else:
        click.echo(rs.stderr.decode('utf-8'))


def print_docker_build_logs(logs) -> None:
    for chunk in logs:
        if 'stream' in chunk:
            for line in chunk['stream'].splitlines():
                click.echo(line)


def build_docker_api_client() -> APIClient:
    docker_url = 'unix://var/run/docker.sock'
    # allows DOCKER_HOST set in the environment to override using the default host
    # docker socket to run the docker commands
    if 'DOCKER_HOST' in os.environ:
        docker_url = os.environ['DOCKER_HOST']
    return APIClient(base_url=docker_url)


def expand_variables(content: str, replace_vars: Dict[str, str]) -> str:
    """
    Does string variable replacement for bash-like variable tokens of the form $VAR or ${VAR}

    :param content: The string to find bash-like variable tokens in
    :param replace_vars: a dictionary of the variable names and values to replace if found
    :return: a string with the bash-like tokens replaced
    """
    for k, v in replace_vars.items():
        if v:
            content = content.replace('$' + k, v)
            content = content.replace('${' + k + '}', v)
    return content


def parse_image_name(image_name: str) -> Tuple[str, str, str]:
    """
    Parses a docker image name into a tuple of registry name, image name and image version

    :param image_name: full docker image name
    :return: tuple of registry name, image name and image version
    """
    if '/' in image_name:
        image_name_parts = image_name.split('/')
        registry_name = image_name_parts[0]
        root_image_name = image_name_parts[1].split(':')[0]
    else:
        registry_name = None
        root_image_name = image_name.split(':')[0]
    image_version = image_name.split(':')[1]
    return registry_name, root_image_name, image_version


def read_env(workspace_metadata: WorkspaceEntryPointMetadata, env_path: Optional[str] = None) -> Dict[str, str]:
    """
    Reads and merges different sets of environment variables that will be leveraged when starting or stopping services

    :param workspace_metadata: metadata for the workspace of the project
    :param env_path: environment variable file path
    :return:
    """
    env = {
        'DISPLAY': os.environ['DISPLAY'],
        'LAB_DOCKER_SOCKET': os.environ['LAB_DOCKER_SOCKET'],
        'LAB_ROOT': os.environ['LAB_ROOT'],
        'LAB_HOME': os.environ['LAB_HOME'],
        'LAB_NETWORK_NAME': os.environ['LAB_NETWORK_NAME'],
        'LAB_VERSION': os.environ['LAB_VERSION'],
        'LAB_PROJECT_WORKSPACE': workspace_metadata.project_path,
        'LAB_PROJECT_WORKSPACE_DATA': workspace_metadata.project_data_path
    }

    set_env_if_exists(env, 'AWS_ACCESS_KEY_ID')
    set_env_if_exists(env, 'AWS_SECRET_ACCESS_KEY')
    set_env_if_exists(env, 'AWS_SESSION_TOKEN')
    set_env_if_exists(env, 'DOCKER_HOST')

    if env_path:
        absolute_env_path = os.path.join(workspace_metadata.project_path, env_path)
        env_file_vars = dotenv_values(absolute_env_path)
        env.update(env_file_vars)
    return env


def set_env_if_exists(env: dict, env_name: str) -> None:
    if env_name in os.environ:
        env[env_name] = os.environ[env_name]


def read_docker_compose_config(workspace_metadata: WorkspaceEntryPointMetadata,
                               compose_config_path: str, env: Dict[str, str]) -> str:
    """
    Reads a docker compose file, replaces any environment variables defined

    :param workspace_metadata: metadata for the workspace of the project
    :param compose_config_path: file system path to a docker compose file
    :param env: environment variable file path
    :return:
    """
    try:
        absolute_config_path = os.path.join(workspace_metadata.project_path, compose_config_path)
        with open(absolute_config_path, 'r') as config:
            prepared_config = expand_variables(config.read(), env)
        rs = subprocess.run(['docker-compose', '-f', '-', 'config'], input=prepared_config.encode(), capture_output=True)
        err = rs.stderr.decode('utf-8')
        if err:
            click.echo(err)
        processed_config = rs.stdout.decode('utf-8')
        return processed_config
    except subprocess.CalledProcessError as e:
        click.echo(e.stderr)
        sys.exit('Error parsing docker-compose.yml')


def exec_compose(compose_cmd: str, compose_config_content: str, env: Optional[Dict[str, str]]=None) -> None:
    """
    Exec a docker compose process by forking the current process allowing the child
    to take over the main process id

    :param compose_cmd: any docker compose command
    :param compose_config_content: a string representation of the contents of a docker compose file
    """
    r, w = os.pipe()
    pid = os.fork()
    if pid == 0:
        # Child process
        os.dup2(r, sys.stdin.fileno())
        os.close(r)
        os.close(w)
        up_command = f'docker-compose -f - {compose_cmd}'.split()
        if env:
            os.execvpe(up_command[0], up_command, env)
        else:
            os.execvp(up_command[0], up_command)
    else:
        # Parent process
        os.close(r)
        os.write(w, bytearray(compose_config_content, 'utf-8'))
        os.close(w)
        pid, status = os.waitpid(pid, 0)


def start_compose(project_name: str, services: str, compose_config_content: str, env: Optional[Dict[str, str]]=None):
    exec_compose(f'-p {project_name} up -d --force-recreate {services}', compose_config_content, env)


def stop_compose(project_name: str, services: str, compose_config_content: str, env: Optional[Dict[str, str]]=None):
    exec_compose(f'-p {project_name} rm --force --stop -v {services}', compose_config_content, env)


def start_services(workspace_metadata: WorkspaceEntryPointMetadata,
                   env_path: Optional[str] = None,
                   compose_config_path: str = 'docker-compose.yml',
                   services: str = '',
                   create_data_directory: bool = True,
                   env_override: Optional[Dict[str, str]]=None) -> None:
    """
    Start services defined in docker compose file

    :param workspace_metadata: metadata for the workspace of the project
    :param env_path: environment variable file path
    :param compose_config_path: a docker compose config file path
    :param services: comma separated list of services to start or omit param to start all
    :param create_data_directory: comma separated list of services to start or omit param to start all
    """
    if create_data_directory and workspace_metadata.project_data_path:
        os.makedirs(workspace_metadata.project_data_path, exist_ok=True)
    elif create_data_directory:
        os.makedirs(os.path.join(workspace_metadata.project_path, 'data'), exist_ok=True)
    env = read_env(workspace_metadata, env_path)
    if env_override:
        env.update(env_override)
    config = read_docker_compose_config(workspace_metadata, compose_config_path, env)
    start_compose(workspace_metadata.project_name, services, config, env)


def stop_services(workspace_metadata: WorkspaceEntryPointMetadata,
                  env_path: Optional[str] = None,
                  compose_config_path: str = 'docker-compose.yml',
                  services: str = '') -> None:
    """
    Stop services running in docker compose

    :param workspace_metadata: metadata for the workspace of the project
    :param env_path: environment variable file path
    :param compose_config_path: a docker compose config file path
    :param services: comma separated list of services to start or omit param to start all
    """
    env = read_env(workspace_metadata, env_path)
    config = read_docker_compose_config(workspace_metadata, compose_config_path, env)
    stop_compose(workspace_metadata.project_name, services, config, env)



from .daemon_info import DockerDaemonInfo
from ..rootless import (
    RootlessContainer,
    LAB_ROOTLESS_DOCKERD,
)
from .run_builder import DockerRunBuilder, UnixUser


__all__ = [
    DockerDaemonInfo,
    RootlessContainer,
    DockerRunBuilder,
    UnixUser,
    LAB_ROOTLESS_DOCKERD
]



import logging
import sys
if sys.version_info < (3, 8):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points
from typing import List, Optional
from click import Command, Context, MultiCommand
from . import CLI_PLUGINS_ENTRY_POINT


logger = logging.getLogger(__name__)


class MetadataEntrypointCommands(MultiCommand):
    """
    Searches installed python packages for CLI entry points
    and loads the specified functions as click Commands
    """

    def list_commands(self, ctx: Context) -> List[str]:
        """
        Generates list of subcommand names from installed packages for cli_plugins entry point

        :param ctx: current click context
        :return: list of subcommand names
        """
        plugin_names = []
        eps = entry_points().get(CLI_PLUGINS_ENTRY_POINT, [])
        for e in eps:
            plugin_names.append(e.name)
        return plugin_names

    def get_command(self, ctx: Context, plugin_name: str) -> Optional[Command]:
        """
        Builds a set of commands for an installed package with a CLI entry point
        and returns a click Command

        :param ctx: current click context
        :param plugin_name: name of the cli plugin in the installed package
        :return: Compiled click Command or None if there was an error
        """
        eps = entry_points().get(CLI_PLUGINS_ENTRY_POINT, [])
        for e in eps:
            if e.name == plugin_name:
                return e.load()


import os
import grp
from typing import Optional
import logging


logger = logging.getLogger(__name__)


def home() -> str:
    """Home directory of the current user

    :return: path to home directory
    """
    return os.environ['HOME']

def username() -> str:
    """Name of the current user

    :return: current user name
    """
    return os.environ['USER']

def user_runtime_tmp_path() -> str:
    return os.environ['XDG_RUNTIME_DIR']

def uid() -> int:
    """UID of the current user

    :return: current user UID
    """
    return os.getuid()

def gid() -> int:
    """GID of the current user

    :return: current user GID
    """
    return os.getgid()

def docker_gid() -> Optional[int]:
    try:
        return grp.getgrnam('docker').gr_gid
    except KeyError:
        logger.debug('Group "docker" does not exist')


def home_subdir(subdir: str) -> str:
    """
    Returns the path to a subdirectory under the user's home directory on the host system.
    :param subdir: Subdirectory (e.g. ".ssh")
    :return: Absolute path to home sub
    """
    return os.path.join(home(), subdir)


#!/usr/bin/env python

import logging
import os
import click

from lab_partner.docker import DockerEnvironment
from lab_partner.rootless import RootlessContainer
from lab_partner.shell import ShellContainer


def loglevel() -> str:
    return os.environ.get('LOGLEVEL', 'INFO').upper()


logging.basicConfig(level=loglevel())
logger = logging.getLogger(__name__)


# DISPLAY = os.environ.get('DISPLAY', '')
# WORKSPACE = os.environ['LAB_WORKSPACE']
# WORKSPACE_DATA = os.path.join(WORKSPACE, 'data')
# NETWORK_NAME = 'lab'


@click.group(invoke_without_command=True)
@click.pass_context
def lab(ctx: click.Context):
    """
    Core Lab Management
    """
    if not ctx.invoked_subcommand:
        ctx.invoke(start)
        ctx.invoke(shell)
    else:
        click.echo(f"I am about to invoke {ctx.invoked_subcommand}")


@lab.command()
@click.option('--debug-rootless', is_flag=True, show_default=True, default=False, help='Turn debugging for Rootlesskit')
@click.option('--debug-bypass4netns', is_flag=True, show_default=True, default=False, help='Turn on debugging for Bypass4netns')
def start(debug_rootless: bool, debug_bypass4netns: bool):
    env = {
        'DOCKERD_ROOTLESS_ROOTLESSKIT_DEBUG': str(debug_rootless).lower(),
        'BYPASS4NETNS_DEBUG': str(debug_bypass4netns).lower()
    }
    rootless = RootlessContainer(DockerEnvironment(env))
    rootless.start()

    # docker_daemon_info = DockerDaemonInfo.build()
    # rootless = RootlessDockerContainer(docker_daemon_info)
    # docker_daemon_info = rootless.start(WORKSPACE)

    # cli_cmd = DockerRunBuilder(f'enclarify/lab-partner-cli:{lab_version()}')
    # cli_cmd.options() \
    #     .with_tty() \
    #     .with_env('LAB_DOCKER_SOCKET', docker_daemon_info.docker_internal_socket()) \
    #     .with_env('LAB_WORKSPACE', os.environ.get('LAB_WORKSPACE')) \
    #     .with_env('LAB_WORKSPACE', WORKSPACE) \
    #     .with_env('LAB_WORKSPACE_DATA', WORKSPACE_DATA) \
    #     .with_env('LAB_NETWORK_NAME', NETWORK_NAME) \
    #     .with_env('LAB_VERSION', lab_version()) \
    #     .with_home_dir_bind_mount('.gitconfig', '/opt/lab/home/.gitconfig') \
    #     .with_home_dir_bind_mount('.vim', '/opt/lab/home/.vim') \
    #     .with_home_dir_bind_mount('.vimrc', '/opt/lab/home/.vimrc') \
    #     .with_home_dir_bind_mount('.actrc', '/opt/lab/home/.actrc') \
    #     .with_home_dir_bind_mount('.aws', '/opt/lab/home/.aws') \
    #     .with_home_dir_bind_mount('.ssh', '/opt/lab/home/.ssh') \
    #     .with_bind_mount(WORKSPACE, WORKSPACE) \
    #     .with_bind_mount('/opt/lab/cicd/artifacts', '/opt/lab/cicd/artifacts') \
    #     .with_bind_mount(docker_daemon_info.docker_socket(), '/var/run/docker.sock') \
    #     .with_workdir(WORKSPACE)

    # cmd = shlex.split(cli_cmd.build())
    # os.execvpe(cmd[0], cmd, {'DOCKER_HOST': f'unix://{docker_daemon_info.docker_socket()}'})


@lab.command()
def stop():
    rootless = RootlessContainer(DockerEnvironment())
    rootless.stop()


@lab.command()
@click.option('--log-level', type=click.Choice(['DEBUG', 'INFO', 'WARN', 'ERROR', 'CRITICAL'], case_sensitive=False), default='INFO')
def shell(log_level: str):
    env = {'LOG_LEVEL': log_level.upper()}
    shell = ShellContainer(RootlessContainer.rootless_env().put(env))
    shell.start()


if __name__ == '__main__':
    lab()



import logging
from lab_partner.docker import DockerComposeClient, DockerEnvironment, LabelFilter
from . import config


logger = logging.getLogger(__name__)


LAB_SHELL_CONTAINER_NAME = 'lab-shell'
LAB_SHELL_COMPOSE_CONFIG_PATH = 'lab_partner.compose_config.shell'


class ShellContainer:
    def __init__(self, env: DockerEnvironment):
        self._container_name = LAB_SHELL_CONTAINER_NAME
        self._docker_client = DockerComposeClient(config.project_name(), LAB_SHELL_COMPOSE_CONFIG_PATH, env)
        self._role_label_filter = LabelFilter(config.lab_role_label(), config.lab_shell_role())

    def start(self) -> None:
        self._docker_client.exec_service(
            LAB_SHELL_CONTAINER_NAME,
            env={
                'LAB_VERSION': config.version(),
                'LAB_WORKSPACE': config.workspace(),
                'LAB_WORKSPACE_DATA': config.workspace_data(),
                'LAB_CICD_ARTIFACT_STORAGE': config.cicd_storage_path(),
                'LAB_ROLE_LABEL': config.lab_role_label(),
                'LAB_SHELL_ROLE': config.lab_shell_role()
            }
        )



#!/usr/bin/env python

import sys
import logging

from click import CommandCollection

from .command_sets.metadata_entrypoint_commands import MetadataEntrypointCommands
from .command_sets.workspace_entrypoint_commands import WorkspaceEntrypointCommands


logger = logging.getLogger(__name__)


plugin_commands = MetadataEntrypointCommands()
workspace_commands = WorkspaceEntrypointCommands()
cli = CommandCollection(sources=[workspace_commands, plugin_commands])


if __name__ == '__main__':
    cli()



import os
from pytest import fixture
from lab_partner.docker_old.run_builder import DockerRunOptions


def assert_only_one(options: DockerRunOptions, option_prefix: str) -> None:
    assert options.build().count(option_prefix) == 1


@fixture(scope="session", autouse=True)
def setup_env():
    os.environ['HOME'] = '/home/me'
    os.environ['USER'] = 'me'
    yield "run test"
    del os.environ['HOME']
    del os.environ['USER']


def test_init_option():
    options = DockerRunOptions()
    option_prefix = '--init'
    options.with_init()
    assert_only_one(options, option_prefix)
    options.with_init()
    assert_only_one(options, option_prefix)


def test_remove_on_exit_option():
    options = DockerRunOptions()
    option_prefix = '--rm'
    options.with_remove_on_exit()
    assert_only_one(options, option_prefix)
    options.with_remove_on_exit()
    assert_only_one(options, option_prefix)


def test_name_option():
    options = DockerRunOptions()
    option_prefix = '--name'
    option_value = 'bob'
    options.with_name(option_value)
    assert_only_one(options, option_prefix)
    option_value = 'sue'
    options.with_name(option_value)
    assert_only_one(options, option_prefix)


def test_hostname_option():
    options = DockerRunOptions()
    option_prefix = '--hostname'
    option_value = 'bob'
    options.with_hostname(option_value)
    assert_only_one(options, option_prefix)
    option_value = 'sue'
    options.with_hostname(option_value)
    assert_only_one(options, option_prefix)


def test_user_option():
    options = DockerRunOptions()
    option_prefix = '--user'
    option_value1 = 1
    option_value2 = 1
    options.with_user(option_value1, option_value2)
    assert_only_one(options, option_prefix)
    option_value1 = 2
    option_value2 = 2
    options.with_user(option_value1, option_value2)
    assert_only_one(options, option_prefix)





#!/usr/local/bin/python

import os
import tomli_w
try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib


template = """
[project]
name = 'dev-test-dependencies'
version = '1.0.0'
dependencies = []
[project.optional-dependencies]
dev = []
test = []
"""

os.mkdir('/tmp/deps')
with open('pyproject.toml', 'rb') as f:
    project_toml = tomllib.load(f)

deps_toml = tomllib.loads(template)
deps_toml['project']['dependencies'] = project_toml['project']['dependencies']
deps_toml['project']['optional-dependencies']['dev'] = project_toml['project']['optional-dependencies']['dev']
deps_toml['project']['optional-dependencies']['test'] = project_toml['project']['optional-dependencies']['test']

with open('/tmp/deps/pyproject.toml', 'wb') as f:
    tomli_w.dump(deps_toml, f)


